/*
    SWWViewer

    An OpenSceneGraph viewer for pyVolution SWW files.
    copyright (C) 2004 Geoscience Australia
*/

#include <osg/Group>
#include <osg/Material>
#include <osg/MatrixTransform>
#include <osg/Notify>
#include <osg/PositionAttitudeTransform>
#include <osg/StateAttribute>

#include <project.h>
#include <SWWReader.h>
#include <bedslope.h>
#include <customviewer.h>
#include <hud.h>
#include <keyboardeventhandler.h>
#include <spotlight.h>
#include <watersurface.h>


// prototypes
osg::Geode* createSky(float radius, char* filename);



int main( int argc, char **argv )
{

    std::cout << argv[0] << std::endl;

    // use an ArgumentParser object to manage the program arguments.
    osg::ArgumentParser arguments(&argc,argv);

    // set up the usage document, in case we need to print out how to use this program.
    arguments.getApplicationUsage()->setDescription(arguments.getApplicationName());
    arguments.getApplicationUsage()->setCommandLineUsage("swwviewer [options] swwfile ...");
   
    // construct the viewer.
    CustomViewer viewer(arguments);

    // set up the value with sensible default event handlers.
    viewer.setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);
    viewer.setClearColor(osg::Vec4(DEF_BACKGROUND_COLOUR));
    viewer.getCamera(0)->getRenderSurface()->setWindowRectangle(550,100,640,480);


    // get details on keyboard and mouse bindings used by the viewer.
    viewer.getUsage(*arguments.getApplicationUsage());

    // if user request help write it out to cout.
    if (arguments.read("-h") || arguments.read("--help"))
    {
        arguments.getApplicationUsage()->write(std::cout);
        return 1;
    }


    // load sww file specified as final argument on command line (static 
    // geometry only, per timestep height field geometry is done in loop below)
    int lastarg = arguments.argc()-1;
    std::string swwfile = arguments.argv()[lastarg];
    arguments.remove(lastarg);
    if( swwfile.substr(swwfile.size()-4,4).find(".sww",0) == -1 )  // ensure filename ends in .sww
    {
        std::cout << "Require last argument be an .sww file ... quitting" << std::endl;
        return 1; 
    }
    SWWReader *sww = new SWWReader(swwfile);
    if (sww->isValid() == false)
    {
        std::cout << "Could not load " << swwfile << " ... quitting" << std::endl;
        return 1;
    }


    // default arguments
    float tps;
    if( !arguments.read("-tps",tps) || tps <= 0.0 ) tps = DEF_TPS;


    // root node
    osg::Group* rootnode = new osg::Group;

    // transform
    osg::PositionAttitudeTransform* model = new osg::PositionAttitudeTransform;
    model->setName("model_transform");

    // enscapsulates OpenGL state
    osg::StateSet* rootStateSet = new osg::StateSet;

    // Bedslope geometry
    BedSlope* bedslope = new BedSlope(sww);

    // Water geometry
    WaterSurface* water = new WaterSurface(sww);

    // Heads Up Display (text overlay)
    HeadsUpDisplay* hud = new HeadsUpDisplay();
    hud->setTitle("pyVolution SWW Viewer");

    // Lighting
    SpotLight* spotlight = new SpotLight(rootStateSet);
    spotlight->setPosition( osg::Vec3(0,0,2) );  // overhead
    spotlight->setSpotAngle( 45.0 );

    // scenegraph hierarchy
    rootnode->setStateSet(rootStateSet);
    rootnode->addChild( hud->get() );
    rootnode->addChild( spotlight->get() );
    rootnode->addChild(model);
    model->addChild( bedslope->get() );
    model->addChild( water->get() );


    // FIXME: should be reasonable default and adjustable
    model->setScale( osg::Vec3(1.0, 1.0, 1.0) );

    // sky
    rootnode->addChild( createSky(10.0, "sky_small.jpg") ); 


    // add model to viewer.
    viewer.setSceneData(rootnode);
 
    // any option left unread are converted into errors to write out later.
    arguments.reportRemainingOptionsAsUnrecognized();
 
    // report any errors if they have occured when parsing the program aguments.
    if (arguments.errors())
    {
        arguments.writeErrorMessages(std::cout);
        return 1;
    }

    // register additional event handler
    KeyboardEventHandler* event_handler = new KeyboardEventHandler(sww->getNumberOfTimesteps(), tps);
    viewer.getEventHandlerList().push_front(event_handler);
 
    // create the windows and run the threads.
    viewer.realize();


    // Initial camera position
    viewer.getTrackball()->setNode( rootnode );
    viewer.getTrackball()->setAutoComputeHomePosition(false);
    viewer.getTrackball()->setHomePosition(
        osg::Vec3d(0,-4,0),  // camera location
        osg::Vec3d(0,0,0),     // camera target
        osg::Vec3d(0,0,1) );   // camera up vector
    viewer.getTrackball()->moveToHome();


    unsigned int timestep = 0;
    while( !viewer.done() )
    {
        // wait for all cull and draw threads to complete.
        viewer.sync();

        // current time
        double time = viewer.getFrameStamp()->getReferenceTime();

        // advance sww frame?
        event_handler->setTime( time );
        if( event_handler->timestepChanged() )
        {
            timestep = event_handler->getTimestep();
            water->setTimeStep(timestep);
            hud->setTime( sww->getTime(timestep) );
        }

        // events
        if( event_handler->toggleWireframe() )
            water->toggleWireframe();
        if( event_handler->toggleCullSetting() )
        {
            sww->toggleCullSetting();
	        water->setTimeStep(timestep);  // forced reload
        }


        // update the scene by traversing with the update visitor 
        viewer.update();
         
        // fire off the cull and draw traversals of the scene.
        viewer.frame();
    }
    
    // wait for all cull and draw threads to complete before exit.
    viewer.sync();

    return 0;
}
